// Lzsa2Decode:
#ifndef FLAT32
//   input   ds:si=inStream, es:di=outStream
//   output  outStream[], ds:si, es:di
	.code16
#define AX	%ax
#define BX	%bx
#define BP	%bp
#define SI	%si
#define DI	%di
#else
//   input   esi=inStream, edi=outStream
//   output  outStream[], ds:esi, es:edi
	.code32
#define AX	%eax
#define BX	%ebx
#define BP	%ebp
#define SI	%esi
#define DI	%edi
#endif

MATCH_RUN_LEN		=	7
LITERALS_RUN_LEN	=	3
MIN_MATCH_SIZE		=	2
MIN_LITERALS_SIZE	=	0

#define PACKED_ONLY			// assume no copy block, optional
//#define PARANOIA			// cover rare cases, optional
 
lzsa2main:
#ifdef PARANOIA
	cld
#endif
#ifdef FLAT16OUT
#define RAW_FORMAT
#endif
#ifndef RAW_FORMAT
# if defined(PARANOIA) && !defined(FLAT32) && !defined(FLAT16)
	xorw	%cx, %cx
	call	normalize
# endif
# ifndef NO_LZSA2_HEADER
	lodsw
	cmpw	$0x9E7B, %ax	// magic
	jne	lzsa2main
	lodsb
	testb	$0x20, %al	// lzsa2
	je	lzsa2main
# endif
lzsa2block:			// uncompress chunk
# if !defined(FLAT32) && !defined(FLAT16)
	xorw	%cx, %cx
	call	normalize
# endif
	lodsw			// block size
	xchgw	%ax, %cx
	lodsb
# ifndef PACKED_ONLY
	orb	%al, %al
	jns	lzsa2compressed
#  if !defined(FLAT32) && !defined(FLAT16OUT)
	movw	%cx, %dx
	movb	$0, %cl
	movb	$0, %dh
copytail:
	call	lzsa1movStr
	xchg	%dx, %cx
	incw	%cx
	loop	copytail
#  else
	movsb			// copy block
copylp:
	movsb			// copy block
	loop	copylp		// handle 64K case
#  endif
	jmp	lzsa2block	// end of block
lzsa2compressed:
	jne	lzsa2chunk	// 64Kb block
# endif
	jcxz	lzsa2quit	// bail if we hit EOD
# if !defined(FLAT16)
	pushw	%cx
	xorw	%cx, %cx
	call	normalize
# define NeedNormalize
	popw	%dx
	addw	%si, %dx
# else
	movw	%si, %dx
	addw	%cx, %dx
# endif
#else
# if !defined(FLAT16)
	xorw	%cx, %cx
	call	normalize
# define NeedNormalize
# endif
#endif
	movb	$0, %bh		// no nibble stored
lzsa2chunk:			// uncompress chunk
	lodsb			// get token XYZ|LL|MMM
	movb	%al, %bl	// keep token in bl
	movw	$LITERALS_RUN_LEN+256*MIN_LITERALS_SIZE, %cx
	shrb	%cl, %al	// shift literals length into place
	call	lzsa2len	// %cl = LITERALS_RUN_LEN
#if defined(NeedNormalize)
	pushw	%bp
	call	lzsa2movLit	// copy %cx literals from %ds:%si to %es:%di
	popw	%bp
#else
	rep	movsb		// copy %cx literals from %ds:%si to %es:%di
#endif
#ifndef RAW_FORMAT
maxsi:
	cmpw	%dx, %si
	jae	lzsa2block	// bail if we hit EOD
#endif
#ifdef FLAT32
	orl	$-1, %eax	// set offset bits 31-8 to 1
#else
	movb	$-1, %ah	// set offset bits 15-8 to 1
#endif
// XYZ
	testb	$0xC0, %bl	// check match offset mode in token (X bit)
	bt	$5, %bx		// move bit 5 to carry
	js	rep_match_or_large_offset
	jne	offset_9_bit
// 00Z 5-bit offset: read a nibble for offset bits 1-4 and use the inverted bit Z of the token as bit 0 of the offset.
//                   set bits 5-15 of the offset to 1.
	call	getByteFromNibbleAndC
	jmp	get_match_length
offset_9_bit:
// 01Z 9-bit offset: read a byte for offset bits 0-7 and use the inverted bit Z for bit 8 of the offset.
//                   set bits 9-15 of the offset to 1.
	sbbb	%cl, %ah	// clear bit 8 if Z bit is clear
	jmp	get_match_length_0

getNibble:
	xorb	$0xF0, %bh	// toggle nibble stored flags
	movb	%bh, %al
	jns	gotnibble
	lodsb
	movb	$0xF0, %bh
	orb	%al, %bh
	shrb	$4, %al
gotnibble:
lzsa2quit:
	ret

rep_match_or_large_offset:
	jpe	rep_match_or_16_bit
//10Z 13-bit offset: read a nibble for offset bits 9-12 and use the inverted bit Z for bit 8 of the offset,
//                   then read a byte for offset bits 0-7. set bits 13-15 of the offset to 1.
//                   substract 512 from the offset to get the final value.
	call	getByteFromNibbleAndC
	subb	$2, %al		// substract 512
	jmp	get_match_length_1
rep_match_or_16_bit:
	jc	repeat_match	// rep-match
//110 16-bit offset: read a byte for offset bits 8-15, then another byte for offset bits 0-7.
	lodsb			// Get 2-byte match offset
get_match_length_1:
	xchgb	%al, %ah
get_match_length_0:
	lodsb			// load match offset bits 0-7
get_match_length:
	xchgw	%ax, %bp	// bp: offset
repeat_match:
//111 repeat offset: reuse the offset value of the previous match command.

	movb	%bl, %al	// %al: original token
	movw	$MATCH_RUN_LEN+256*MIN_MATCH_SIZE, %cx
	call	lzsa2len
#ifdef RAW_FORMAT
	jcxz	lzsa2quit	// bail if we hit EOD
#endif
#if !defined(FLAT32) && !defined(FLAT16OUT)
	pushw	%ds
	pushw	%si
	movw	%di, %si
	addw	%bp, %si
	movw	%es, %ax
	jc	axok
	subb	$0x10, %ah
axok:
.macro norm	reg
	movw	%si, \reg
	subw	%si, %dx
	andw	$0xF, %si
	addw	%si, %dx
	shrw	$4, \reg
	addw	\reg, %ax
	movw	%ax, %ds
	movw	%di, \reg
	andw	$0xF, %di
	shrw	$4, \reg
	movw	%es, %ax
	addw	\reg, %ax
	movw	%ax, %es
.endm 
	pushw	%bp
	pushw	%dx
# if defined(NeedNormalize) || defined(PARANOIA)
	call	lzsa2movStr	// copy string
# else
	norm	%bp
	rep movsb
# endif
	popw	%dx
	popw	%bp
	popw	%si
	popw	%ds
#else
	xchg	AX, SI		// save %si	
	lea	(BP,DI), SI
	rep movsb	%es:(SI), %es:(DI)
	xchg	AX, SI		// restore %si	
#endif
	jmp	lzsa2chunk

getByteFromNibbleAndC:
	pushfw
	call	getNibble	// get nibble for offset bits 0-3
	popfw
	rclb	$1, %al
	xorb	$0xE1, %al	// set offset bits 7-5 to 1
	ret

lzsa2len:			// get length in %cx
	andb	%cl, %al
	cbw			// clear %ah
	cmpb	%al, %cl
	jne	lzsa2minNumber	// S=0-2, L=0-6
	call	getNibble
	cmp	$0xF, %al
	jne	lzsa2noExtraByte
	addb	%al, %ch
	lodsb
lzsa2noExtraByte:
	addb	%cl, %ch
lzsa2minNumber:
	addb	%ch, %al
	jnc	lzsa2gotNumber  // 0-255
#if 0
	je	lzsa2BigNumber
	movb	%al, %ah	// S=256-767, L=256-1791
	lodsb	
	.byte	0xB1		// mask lodsw with movb $0xAD, %cl
lzsa2BigNumber:
#endif
	lodsw			// 0-65535
lzsa2gotNumber:
	xchgw	%ax, %cx
	ret

#if defined(NeedNormalize) || defined(PARANOIA)
# if defined(PARANOIA)
lzsa2movlp:
	decw	%ch
	rep movsb
	incw	%ch
# endif
normalize:
lzsa2movLit:
	movw	%ds, %ax
lzsa2movStr:
	norm	%bp
# if defined(PARANOIA)
	cmpb	$0xFF, %ch	// catch FFFX case
	je	lzsa2movlp
# endif
	rep movsb
	ret
#endif
